@@ -2,7 +2,7 @@ package RedisDB;
 
 use strict;
 use warnings;
-our $VERSION = "2.38";
+our $VERSION = "2.39";
 $VERSION = eval $VERSION;
 
 use RedisDB::Error;
@@ -971,13 +971,19 @@ sub cluster_info {
 
 sub cluster_nodes {
     my $self = shift;
+
     my $list = $self->execute(qw(CLUSTER NODES));
+    return $list if ref $list =~ /^RedisDB::Error/;
+
     my @nodes;
     for ( split /^/, $list ) {
         my ( $node_id, $addr, $flags, $master_id, $ping, $pong, $state, @slots ) =
           split / /;
         my %flags = map { $_ => 1 } split /,/, $flags;
         my ( $host, $port ) = split /:/, $addr;
+        unless ($host) {
+            $host = $self->{host}, $addr = "$self->{host}:$port",
+        }
         my $node = {
             node_id            => $node_id,
             address            => $addr,
@@ -992,6 +998,7 @@ sub cluster_nodes {
         };
         push @nodes, $node;
     }
+
     return \@nodes;
 }
 
