@@ -2,7 +2,7 @@ package RedisDB::Cluster;
 
 use strict;
 use warnings;
-our $VERSION = "2.38";
+our $VERSION = "2.39";
 $VERSION = eval $VERSION;
 
 use Carp;
@@ -61,32 +61,39 @@ sub _initialize_slots {
     }
 
     my %new_nodes;
-    my @new_nodes;
+    my $new_nodes;
     for my $node ( @{ $self->{_nodes} } ) {
-        my $redis = RedisDB->new(
-            host        => $node->{host},
-            port        => $node->{port},
-            raise_error => 0,
-        );
+        my $redis = _connect_to_node( $self, $node );
+        next unless $redis;
+
+        my $nodes = $redis->cluster_nodes;
+        next if ref $nodes =~ /^RedisDB::Error/;
+        $new_nodes = $nodes;
+        for (@$nodes) {
+            $new_nodes{"$_->{host}:$_->{port}"}++;
+        }
+
         my $slots = $redis->cluster('SLOTS');
-        next if ref $slots =~ /^RedisDB::Error/;
+        confess "got an error trying retrieve a list of cluster slots: $slots"
+          if ref $slots =~ /^RedisDB::Error/;
         for (@$slots) {
             my ( $ip, $port ) = @{ $_->[2] };
             my $host_id = "$ip:$port";
-            push @new_nodes,
-              {
-                host => $ip,
-                port => $port,
-              } unless $new_nodes{$host_id}++;
             for ( $_->[0] .. $_->[1] ) {
                 $self->{_slot}[$_] = $host_id;
             }
         }
         last;
     }
-    $self->{_nodes} = \@new_nodes;
-    unless ( @{ $self->{_nodes} } ) {
-        confess "couldn't get cluster slots";
+
+    unless ($new_nodes and @$new_nodes) {
+        confess "couldn't get list of cluster nodes";
+    }
+    $self->{_nodes} = $new_nodes;
+
+    # close connections to nodes that are not in cluster
+    for ( keys %{ $self->{_connection} } ) {
+        delete $self->{_connection}{$_} unless $new_nodes{$_};
     }
 
     return;
@@ -117,16 +124,26 @@ sub execute {
         my $redis = $self->{_connection}{$host_id};
         unless ($redis) {
             my ( $host, $port ) = split /:/, $host_id;
-            $redis = $self->{_connection}{$host_id} = RedisDB->new(
-                host        => $host,
-                port        => $port,
-                raise_error => 0,
+            $redis = _connect_to_node(
+                $self,
+                {
+                    host => $host,
+                    port => $port
+                }
             );
         }
 
-        $redis->asking(RedisDB::IGNORE_REPLY) if $asking;
-        $asking = 0;
-        my $res = $redis->execute( $command, $key, @_ );
+        my $res;
+        if ($redis) {
+            $redis->asking(RedisDB::IGNORE_REPLY) if $asking;
+            $asking = 0;
+            $res = $redis->execute( $command, $key, @_ );
+        }
+        else {
+            $res = RedisDB::Error::DISCONNECTED->new(
+                "Couldn't connect to redis server at $host_id");
+        }
+
         if ( ref $res eq 'RedisDB::Error::MOVED' ) {
             if ( $res->{slot} ne $slot ) {
                 confess
@@ -144,7 +161,7 @@ sub execute {
             next;
         }
         elsif ( ref $res eq 'RedisDB::Error::DISCONNECTED' ) {
-            warn "connection to $host_id lost" if $DEBUG;
+            warn "$res" if $DEBUG;
             delete $self->{_connection}{$host_id};
             usleep 100_000;
             if ( $last_connection and $last_connection eq $host_id ) {
@@ -170,6 +187,90 @@ sub execute {
         "Couldn't send command after 10 attempts");
 }
 
+=head2 $self->add_new_node($address[, $master_id])
+
+attach node with the specified I<$address> to the cluster. If I<$master_id> is
+specified, the new node is configured as a replica of the master with the
+specified ID, otherwise it will be a master node itself. Address should be
+specified as a hash containing I<host> and I<port> elements.
+
+=cut
+
+sub add_new_node {
+    my ( $self, $addr, $master_id ) = @_;
+    $addr = _ensure_hash_address($addr);
+
+    my $redis = _connect_to_node( $self, $addr );
+    my $ok;
+    for my $node ( @{ $self->{_nodes} } ) {
+        $redis->cluster( 'MEET', $node->{host}, $node->{port},
+            sub { $ok++ if not ref $_[1] and $_[1] eq 'OK'; warn $_[1] if ref $_[1]; }
+        );
+    }
+    $redis->mainloop;
+    croak "failed to attach node to cluster" unless $ok;
+
+    if ($master_id) {
+        my $attempt = 0;
+        my $nodes   = $redis->cluster_nodes;
+        while ( not grep { $_->{node_id} eq $master_id } @$nodes ) {
+            croak "failed to start replication from $master_id - node is not present"
+              if $attempt++ >= 10;
+            usleep 100_000 * $attempt;
+            $nodes = $redis->cluster_nodes;
+        }
+        my $res = $redis->cluster( 'REPLICATE', $master_id );
+        croak $res if ref $res =~ /^RedisDB::Error/;
+    }
+
+    return 'OK';
+}
+
+sub _ensure_hash_address {
+    my $addr = shift;
+    unless ( ref $addr eq 'HASH' ) {
+        my ( $host, $port ) = split /:/, $addr;
+        croak "invalid address spec: $addr" unless $host and $port;
+        $addr = {
+            host => $host,
+            port => $port
+        };
+    }
+    return $addr;
+}
+
+sub _connect_to_node {
+    my ( $self, $node ) = @_;
+    my $host_key = "$node->{host}:$node->{port}";
+    unless ( $self->{_connection}{$host_key} ) {
+        my $redis = RedisDB->new(
+            host        => $node->{host},
+            port        => $node->{port},
+            raise_error => 0,
+        );
+        $self->{_connection}{$host_key} = $redis->{_socket} ? $redis : undef;
+    }
+    return $self->{_connection}{$host_key};
+}
+
+=head2 $self->random_connection
+
+return RedisDB object that is connected to random node of the cluster.
+
+=cut
+
+sub random_connection {
+    my $self = shift;
+    my ($connection) = values %{ $self->{_connection} };
+    unless ($connection) {
+        for ( @{ $self->{_nodes} } ) {
+            $connection = _connect_to_node( $self, $_ );
+            last if $connection;
+        }
+    }
+    return $connection;
+}
+
 my @crc16tab = (
     0x0000, 0x1021, 0x2042, 0x3063, 0x4084, 0x50a5, 0x60c6, 0x70e7,
     0x8108, 0x9129, 0xa14a, 0xb16b, 0xc18c, 0xd1ad, 0xe1ce, 0xf1ef,
